#include "pch.h"
#include "VirtualSurfaceImageSourceRenderer.h"

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::UI::Xaml::Media::Imaging;

VirtualSurfaceImageSourceRenderer::VirtualSurfaceImageSourceRenderer(VirtualSurfaceImageSource^ vsis) : m_refCount(0),
                                                                                                        m_dpi(96)
{
    IInspectable* vsisInspectable = reinterpret_cast<IInspectable*>(vsis);
    vsisInspectable->QueryInterface(__uuidof(IVirtualSurfaceImageSourceNative), (void **) &m_vsisNative);
    m_vsisNative->RegisterForUpdatesNeeded(this);
}

void VirtualSurfaceImageSourceRenderer::Initialize()
{
    // Virtual method
    CreateDeviceResources();
}

void VirtualSurfaceImageSourceRenderer::CreateDeviceResources()
{
    // Required for compatibility with Direct2D.
    UINT creationFlags = D3D11_CREATE_DEVICE_BGRA_SUPPORT;

#if defined(_DEBUG)
    creationFlags |= D3D11_CREATE_DEVICE_DEBUG;
#endif

    // Supported hardware features
    D3D_FEATURE_LEVEL featureLevels [] =
    {
        D3D_FEATURE_LEVEL_11_1,
        D3D_FEATURE_LEVEL_11_0,
        D3D_FEATURE_LEVEL_10_1,
        D3D_FEATURE_LEVEL_10_0,
        D3D_FEATURE_LEVEL_9_3,
        D3D_FEATURE_LEVEL_9_2,
        D3D_FEATURE_LEVEL_9_1
    };

    // Create the Direct3D 11 API device object and a corresponding context.
    ComPtr<ID3D11Device> device;
    ComPtr<ID3D11DeviceContext> context;
    
    HRESULT hr = D3D11CreateDevice(
        nullptr,					// Specify nullptr to use the default adapter.
        D3D_DRIVER_TYPE_HARDWARE,	// Create a device using the hardware graphics driver.
        0,							// Should be 0 unless the driver is D3D_DRIVER_TYPE_SOFTWARE.
        creationFlags,				// Set debug and Direct2D compatibility flags.
        featureLevels,				// List of feature levels this app can support.
        ARRAYSIZE(featureLevels),	// Size of the list above.
        D3D11_SDK_VERSION,			// Always set this to D3D11_SDK_VERSION for Windows Store apps.
        &device,					// Returns the Direct3D device created.
        nullptr,                    // Returns feature level of device created.
        &context					// Returns the device immediate context.
        );

    if (FAILED(hr))
        throw ref new COMException(hr, "D3D11CreateDevice failure");

    hr = device.As(&m_d3dDevice);

    if (FAILED(hr))
        throw ref new COMException(hr, "d3dDevice.As failure");

    hr = context.As(&m_d3dContext);

    if (FAILED(hr))
        throw ref new COMException(hr, "d3dContext.As failure");

    ComPtr<IDXGIDevice> dxgiDevice;
    hr = device.As(&dxgiDevice);

    if (FAILED(hr))
        throw ref new COMException(hr, "ID3D11Device.As(IDXGIDevice) failure");

    hr = D2D1CreateDevice(dxgiDevice.Get(), nullptr, &m_d2dDevice);

    if (FAILED(hr))
        throw ref new COMException(hr, "D2D1CreateDevice failure");

    hr = m_d2dDevice->CreateDeviceContext(D2D1_DEVICE_CONTEXT_OPTIONS_NONE, &m_d2dContext);

    if (FAILED(hr))
        throw ref new COMException(hr, "d2dDevice->CreateDeviceContext failure");

    hr = m_vsisNative->SetDevice(dxgiDevice.Get());

    if (FAILED(hr))
        throw ref new COMException(hr, "vsisNative->SetDevice failure");
}

void VirtualSurfaceImageSourceRenderer::SetDpi(float dpi)
{
    m_dpi = dpi;
    m_d2dContext->SetDpi(dpi, dpi);
}

ULONG STDMETHODCALLTYPE VirtualSurfaceImageSourceRenderer::AddRef()
{
    m_refCount++;
    return m_refCount;
}

ULONG STDMETHODCALLTYPE VirtualSurfaceImageSourceRenderer::Release()
{
    m_refCount--;
    ULONG newCount = m_refCount;

    if (m_refCount == 0)
        delete this;

    return newCount;
}

HRESULT STDMETHODCALLTYPE VirtualSurfaceImageSourceRenderer::QueryInterface(REFIID riid, void** ppOutput)
{
    *ppOutput = nullptr;
    HRESULT hr = S_OK;

    if (riid == __uuidof(IVirtualSurfaceUpdatesCallbackNative))
    {
        *ppOutput = static_cast<IVirtualSurfaceUpdatesCallbackNative*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IUnknown))
    {
        *ppOutput = this;
        AddRef();
    }
    else
    {
        hr = E_NOINTERFACE;
    }
    return hr;
}

HRESULT STDMETHODCALLTYPE VirtualSurfaceImageSourceRenderer::UpdatesNeeded()
{
    // Get the update rectangles
    ULONG count;
    HRESULT hr = m_vsisNative->GetUpdateRectCount(&count);

    if (FAILED(hr))
        return hr;

    RECT * updates = new RECT[count];
    hr = m_vsisNative->GetUpdateRects(updates, count);

    if (FAILED(hr))
        return hr;

    for (ULONG i = 0; i < count; i++)
        Draw(updates[i]);

    return S_OK;
}

void VirtualSurfaceImageSourceRenderer::Draw(RECT updateRect)
{
    POINT offset;
    ComPtr<IDXGISurface> surface;

    HRESULT hr = m_vsisNative->BeginDraw(updateRect, &surface, &offset);

    if (hr == DXGI_ERROR_DEVICE_REMOVED || hr == DXGI_ERROR_DEVICE_RESET)
    {
        // Recreate device resources and try again
        CreateDeviceResources();
        hr = m_vsisNative->BeginDraw(updateRect, &surface, &offset);
    }

    if (hr != S_OK && hr != DXGI_ERROR_DEVICE_REMOVED && hr != DXGI_ERROR_DEVICE_RESET)
    {
        throw ref new COMException(hr, "vsisNative->BeginDraw failure");
    }

    // Convert offset and updateRect to DIUs
    float dpiAdj = 96 / m_dpi;
    D2D_POINT_2F fOffset = Point2F(dpiAdj * offset.x, dpiAdj * offset.y);
    D2D_RECT_F fUpdateRect = RectF(dpiAdj * updateRect.left,
                                   dpiAdj * updateRect.top,
                                   dpiAdj * updateRect.right,
                                   dpiAdj * updateRect.bottom);

    // Create render target
    ComPtr<ID2D1Bitmap1> bitmap;
    hr = m_d2dContext->CreateBitmapFromDxgiSurface(surface.Get(), nullptr, &bitmap);

    if (FAILED(hr))
        throw ref new COMException(hr, "d2dContext->CreateBitmapFromDxgiSurface failure");

    m_d2dContext->SetTarget(bitmap.Get());

    // Begin drawing with the Direct2D context
    m_d2dContext->BeginDraw();

    // Clip to the update rectangle and translate to offset
    D2D1_RECT_F clipRect = RectF(fOffset.x, 
                                 fOffset.y,
                                 fOffset.x + fUpdateRect.right - fUpdateRect.left,
                                 fOffset.y + fUpdateRect.bottom - fUpdateRect.top);

    m_d2dContext->PushAxisAlignedClip(clipRect, D2D1_ANTIALIAS_MODE_ALIASED);
    m_d2dContext->SetTransform(Matrix3x2F::Translation(fOffset.x, fOffset.y));

    // Virtual method
    Render(fUpdateRect);

    m_d2dContext->SetTransform(IdentityMatrix());
    m_d2dContext->PopAxisAlignedClip();

    hr = m_d2dContext->EndDraw();

    if (FAILED(hr))
        throw ref new COMException(hr, "d2dContext->EndDraw failure");

    m_d2dContext->SetTarget(nullptr);
    hr = m_vsisNative->EndDraw();

    if (FAILED(hr))
         throw ref new COMException(hr, "vsisNative->EndDraw failure");
}
