#include "pch.h"
#include "AliceParagraphGenerator.h"

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;

AliceParagraphGenerator::AliceParagraphGenerator()
{
}

void AliceParagraphGenerator::SetText(std::vector<std::wstring> lines)
{
    // Create an IDWriteFactory for use during this method
    ComPtr<IDWriteFactory> dwriteFactory;
    HRESULT hr = DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, 
                                     __uuidof(IDWriteFactory), 
                                     &dwriteFactory);
    if (FAILED(hr))
        throw ref new COMException(hr, "DWriteCreateFactory failure");

    // Create a basic IDWriteTextFormat for all text
    ComPtr<IDWriteTextFormat> textFormat;
    hr = dwriteFactory->CreateTextFormat(L"Century Schoolbook",
                                         nullptr,
                                         DWRITE_FONT_WEIGHT_NORMAL,
                                         DWRITE_FONT_STYLE_NORMAL,
                                         DWRITE_FONT_STRETCH_NORMAL,
                                         24.0f,
                                         L"en-US",
                                         &textFormat);

    if (FAILED(hr))
        throw ref new COMException(hr, "dwriteFactory->CreateTextFormat failure");

    // Create first-line indent object
    ComPtr<IDWriteInlineObject> halfInchIndent = new FirstLineIndent(48);

    // Clear out any earlier paragraphs
    m_paragraphs.clear();

    // Loop through the lines
    for (unsigned int index = 0; index < lines.size(); index++)
    {
        // Set the text with an initial character for first-line indenting
        Paragraph paragraph;
        paragraph.Text = L"\x200B";
        paragraph.Text.append(lines.at(index).c_str());

        // Create the IDWriteTextLayout object
        HRESULT hr = dwriteFactory->CreateTextLayout(paragraph.Text.c_str(),
                                                     paragraph.Text.length(),
                                                     textFormat.Get(),
                                                     0, 0,
                                                     &paragraph.TextLayout);
        if (FAILED(hr))
            throw ref new COMException(hr, "dwriteFactory->CreateTextLayout failure");

        // Center first two lines
        if (index == 0 || index == 1)
        {
            paragraph.TextLayout->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_CENTER);
            paragraph.SpaceAfter = 36;

            if (FAILED(hr))
               throw ref new COMException(hr, "textLayout->SetInlineObject");
        }
        else
        {
            // Justify the others
            HRESULT hr = paragraph.TextLayout->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_JUSTIFIED);

            if (FAILED(hr))
                throw ref new COMException(hr, "textLayout->SetTextAlignment");

            // Indent by a half inch
            DWRITE_TEXT_RANGE textRange = { 0, 1 };
            hr = paragraph.TextLayout->SetInlineObject(halfInchIndent.Get(), textRange);

            if (FAILED(hr))
                throw ref new COMException(hr, "textLayout->SetInlineObject");

            // A little different paragraph spacing for the vers
            bool noSpace = index == 50 || index == 55 || index == 56;
            paragraph.SpaceAfter = noSpace ? 0.0f : 12.0f;
        }
        m_paragraphs.push_back(paragraph);
    }

    // Italicized words
    SetItalicText(3, L"plenty");
    SetItalicText(9, L"your");
    SetItalicText(12, L"said");
    SetItalicText(21, L"is");
    SetItalicText(25, L"best");
    SetItalicText(27, L"best");
    SetItalicText(29, L"your");
    SetItalicText(31, L"mine");
    SetItalicText(40, L"it");
    SetItalicText(40, L"him");
    SetItalicText(48, L"you");
    SetItalicText(49, L"he", 4);
    SetItalicText(50);
    SetItalicText(51);
    SetItalicText(55);
    SetItalicText(56);
    SetItalicText(57);
    SetItalicText(58, L"Twinkle, twinkle, twinkle, twinkle");
    SetItalicText(78, L"very");
    SetItalicText(82, L"less");
    SetItalicText(82, L"more");
    SetItalicText(83, L"your");
    SetItalicText(88, L"one");
    SetItalicText(96, L"in");
    SetItalicText(107, L"there");
    SetItalicText(109, L"then", 2);

    // Different font sizes
    float baseFontSize = textFormat->GetFontSize();
    SetFontSize(0, 0, 13, 1.25f * baseFontSize);
    SetFontSize(1, 0, 17, 0.80f * baseFontSize);
    SetFontSize(2, 2, 4, 0.65f * baseFontSize);
}

void AliceParagraphGenerator::SetItalicText(int lineNum, const wchar_t * word, int occurrence)
{
    Paragraph paragraph = m_paragraphs.at(lineNum);
    DWRITE_TEXT_RANGE textRange;

    if (word == nullptr)
    {
        textRange.startPosition = 0;
        textRange.length = wcslen(paragraph.Text.c_str());
    }
    else
    {
        int index = -1;

        for (int rep = 0; rep < occurrence; rep++)
            index = paragraph.Text.find(word, index + 1);

        textRange.startPosition = index;
        textRange.length = wcslen(word);
    }
    HRESULT hr = paragraph.TextLayout->SetFontStyle(DWRITE_FONT_STYLE_ITALIC, textRange);

    if (FAILED(hr))
        throw ref new COMException(hr, "textLayout->SetFontStyle failure");
}

void AliceParagraphGenerator::SetFontSize(int lineNum, int start, int length, float size)
{
    Paragraph paragraph = m_paragraphs.at(lineNum);
    DWRITE_TEXT_RANGE textRange = { start, length };
    HRESULT hr = paragraph.TextLayout->SetFontSize(size, textRange);

    if (FAILED(hr))
        throw ref new COMException(hr, "textLayout->SetFontSize failure");
}

float AliceParagraphGenerator::SetWidth(float width)
{
    if (width <= 0)
        return 0;

    float totalHeight = 0;

    for (Paragraph& paragraph : m_paragraphs)
    {
        HRESULT hr = paragraph.TextLayout->SetMaxWidth(width);

        if (FAILED(hr))
            throw ref new COMException(hr, "textLayout->SetMaxWidth failure");

        hr = paragraph.TextLayout->SetMaxHeight(FloatMax());

        if (FAILED(hr))
            throw ref new COMException(hr, "textLayout->SetMaxHeight failure");

        DWRITE_TEXT_METRICS textMetrics;
        hr = paragraph.TextLayout->GetMetrics(&textMetrics);

        if (FAILED(hr))
            throw ref new COMException(hr, "textLayout->GetMetrics failure");

        paragraph.TextHeight = textMetrics.height;
        totalHeight += paragraph.TextHeight + paragraph.SpaceAfter;
    }
    return totalHeight;
}

std::vector<Paragraph> AliceParagraphGenerator::GetParagraphs()
{
    return m_paragraphs;
}

int AliceParagraphGenerator::ParagraphCount()
{
    return m_paragraphs.size();
}

